# -*- coding: utf-8 -*-

""" #+begin_org
* ~[Summary]~ :: A =CS-Lib= for InMail Abstracted Accessible Service (aas) Offline Imap.
#+end_org """

####+BEGIN: b:py3:cs:file/dblockControls :classification "cs-u"
""" #+begin_org
* [[elisp:(org-cycle)][| /Control Parameters Of This File/ |]] :: dblk ctrls classifications=cs-u
#+BEGIN_SRC emacs-lisp
(setq-local b:dblockControls t) ; (setq-local b:dblockControls nil)
(put 'b:dblockControls 'py3:cs:Classification "cs-u") ; one of cs-mu, cs-u, cs-lib, bpf-lib, pyLibPure
#+END_SRC
#+RESULTS:
: cs-u
#+end_org """
####+END:

####+BEGIN: b:prog:file/proclamations :outLevel 1
""" #+begin_org
* *[[elisp:(org-cycle)][| Proclamations |]]* :: Libre-Halaal Software --- Part Of BISOS ---  Poly-COMEEGA Format.
** This is Libre-Halaal Software. © Neda Communications, Inc. Subject to AGPL.
** It is part of BISOS (ByStar Internet Services OS)
** Best read and edited  with Blee in Poly-COMEEGA (Polymode Colaborative Org-Mode Enhance Emacs Generalized Authorship)
#+end_org """
####+END:

####+BEGIN: b:prog:file/particulars :authors ("./inserts/authors-mb.org")
""" #+begin_org
* *[[elisp:(org-cycle)][| Particulars |]]* :: Authors, version
** This File: /bisos/git/auth/bxRepos/bisos-pip/qmail/py3/bisos/qmail/qmailRemote.py
** Authors: Mohsen BANAN, http://mohsen.banan.1.byname.net/contact
#+end_org """
####+END:

####+BEGIN: b:python:file/particulars-csInfo :status "inUse"
""" #+begin_org
* *[[elisp:(org-cycle)][| Particulars-csInfo |]]*
#+end_org """
import typing

csInfo: typing.Dict[str, typing.Any] = { 'moduleName': ['qmailRemote'], }
csInfo['version'] = '202212122335'
csInfo['status']  = 'inUse'
csInfo['panel'] = 'qmailRemote-Panel.org'
csInfo['groupingType'] = 'IcmGroupingType-pkged'
csInfo['cmndParts'] = 'IcmCmndParts[common] IcmCmndParts[param]'
####+END:

csInfo['description'] = """ #+begin_org
* /[[elisp:(org-cycle)][| Description |]]/ :: [[file:/bisos/git/auth/bxRepos/blee-binders/bisos-core/COMEEGA/_nodeBase_/fullUsagePanel-en.org][BISOS COMEEGA Panel]]
Module description comes here.
** Relevant Panels:
** Status: In use with blee3
** /[[elisp:(org-cycle)][| Planned Improvements |]]/ :
*** TODO complete fileName in particulars.
#+end_org """

####+BEGIN: b:prog:file/orgTopControls :outLevel 1
""" #+begin_org
* [[elisp:(org-cycle)][| Controls |]] :: [[elisp:(delete-other-windows)][(1)]] | [[elisp:(show-all)][Show-All]]  [[elisp:(org-shifttab)][Overview]]  [[elisp:(progn (org-shifttab) (org-content))][Content]] | [[elisp:(blee:ppmm:org-mode-toggle)][Nat]] | [[elisp:(bx:org:run-me)][Run]] | [[elisp:(bx:org:run-me-eml)][RunEml]] | [[elisp:(progn (save-buffer) (kill-buffer))][S&Q]]  [[elisp:(save-buffer)][Save]]  [[elisp:(kill-buffer)][Quit]] [[elisp:(org-cycle)][| ]]
** /Version Control/ ::  [[elisp:(call-interactively (quote cvs-update))][cvs-update]]  [[elisp:(vc-update)][vc-update]] | [[elisp:(bx:org:agenda:this-file-otherWin)][Agenda-List]]  [[elisp:(bx:org:todo:this-file-otherWin)][ToDo-List]]
#+end_org """
####+END:

####+BEGIN: b:python:file/workbench :outLevel 1
""" #+begin_org
* [[elisp:(org-cycle)][| Workbench |]] :: [[elisp:(python-check (format "/bisos/venv/py3/bisos3/bin/python -m pyclbr %s" (bx:buf-fname))))][pyclbr]] || [[elisp:(python-check (format "/bisos/venv/py3/bisos3/bin/python -m pydoc ./%s" (bx:buf-fname))))][pydoc]] || [[elisp:(python-check (format "/bisos/pipx/bin/pyflakes %s" (bx:buf-fname)))][pyflakes]] | [[elisp:(python-check (format "/bisos/pipx/bin/pychecker %s" (bx:buf-fname))))][pychecker (executes)]] | [[elisp:(python-check (format "/bisos/pipx/bin/pycodestyle %s" (bx:buf-fname))))][pycodestyle]] | [[elisp:(python-check (format "/bisos/pipx/bin/flake8 %s" (bx:buf-fname))))][flake8]] | [[elisp:(python-check (format "/bisos/pipx/bin/pylint %s" (bx:buf-fname))))][pylint]]  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:


####+BEGIN: b:py3:cs:orgItem/basic :type "=PyImports= " :title "*Py Library IMPORTS*" :comment "-- with classification based framework/imports"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  =PyImports=  [[elisp:(outline-show-subtree+toggle)][||]] *Py Library IMPORTS* -- with classification based framework/imports  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:

####+BEGIN: b:py3:cs:framework/imports :basedOn "classification"
""" #+begin_org
** Imports Based On Classification=cs-u
#+end_org """
from bisos import b
from bisos.b import cs
from bisos.b import b_io

import collections
####+END:

import collections

import collections
import pathlib
import os
#import shutil

from bisos.bpo import bpoRunBases
from bisos.bpo import bpo

from bisos.common import csParam

#from bisos.marmee import aasInMailControl
from bisos.marmee import aasInMailFps
from bisos.marmee import aasOutMailFps
from bisos.marmee import aasMailFps
from bisos.marmee import gmailOauth2
from bisos.marmee import x822Lib

from bisos.qmail import qmailRemoteWrapper

import sys
import pwd

from email.message import EmailMessage
from email.parser import Parser
from email.utils import parseaddr, getaddresses
from os.path import expanduser
from configparser import ConfigParser
from collections import namedtuple

import smtplib
import http
import urllib
import urllib.request
import requests
import json
import base64
import enum

import tempfile

####+BEGIN: blee:bxPanel:foldingSection :outLevel 0 :sep nil :title "CmndSvcs" :anchor ""  :extraInfo "Command Services Section"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*     [[elisp:(outline-show-subtree+toggle)][| _CmndSvcs_: |]]  Command Services Section  [[elisp:(org-shifttab)][<)]] E|
#+end_org """
####+END:


####+BEGIN: bx:cs:py3:section :title "CS-Lib Examples"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  /Section/    [[elisp:(outline-show-subtree+toggle)][||]] *CS-Lib Examples*  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:

####+BEGIN: b:py3:cs:func/typing :funcName "examples_csu" :funcType "eType" :retType "" :deco "default" :argsList ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  F-T-eType  [[elisp:(outline-show-subtree+toggle)][||]] /examples_csu/  deco=default  [[elisp:(org-cycle)][| ]]
#+end_org """
@cs.track(fnLoc=True, fnEntry=True, fnExit=True)
def examples_csu(
####+END:
        sectionTitle: typing.AnyStr = "",
) -> None:
    """ #+begin_org
** [[elisp:(org-cycle)][| *DocStr | ] Examples of Service Access Instance Commands.
    #+end_org """

    def cpsInit(): return collections.OrderedDict()
    def menuItem(verbosity, **kwArgs): cs.examples.cmndInsert(cmndName, cps, cmndArgs, verbosity=verbosity, **kwArgs) # 'little' or 'none'
    def execLineEx(cmndStr): cs.examples.execInsert(execLine=cmndStr)

    if sectionTitle == "default":
        cs.examples.menuChapter('*Qmail Inject From Stdin*')

    myName = cs.G.icmMyName()
    execLineEx(f"""{myName} -- -n < ~/example.mail""")
    execLineEx(f"""ls -t /tmp/* | head -20 | grep qmail-inject- | head -1""")
    execLineEx(f"""sudo cat $(ls -t /tmp/* | head -20 | grep qmail-inject- | head -1)""")

    cmndName = "qmailInjectCmnd" ;  cmndArgs = "-- -n"
    cps=cpsInit(); cps['runMode'] = "runDebug"
    menuItem(verbosity='none', icmWrapper="cat ~/example.mail | ")
    menuItem(verbosity='full', icmWrapper="cat ~/example.mail | ")

    cmndName = "monolithicSend" ;  cmndArgs = "-- -n"
    cps=cpsInit(); cps['runMode'] = "runDebug"
    menuItem(verbosity='none', icmWrapper="cat ~/example.mail | ")
    menuItem(verbosity='full', icmWrapper="cat ~/example.mail | ")


####+BEGIN: bx:cs:py3:section :title "CS-Params  --- Place Holder, Empty"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  /Section/    [[elisp:(outline-show-subtree+toggle)][||]] *CS-Params  --- Place Holder, Empty*  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:

####+BEGIN: blee:bxPanel:foldingSection :outLevel 0 :sep nil :title "Functions" :anchor ""  :extraInfo "maildrop stdout/update"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*     [[elisp:(outline-show-subtree+toggle)][| _Functions_: |]]  maildrop stdout/update  [[elisp:(org-shifttab)][<)]] E|
#+end_org """
####+END:

####+BEGIN: bx:dblock:python:enum :enumName "MetaQmailQueueOfQmailInject" :comment "Enum Values: "
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Enum       [[elisp:(outline-show-subtree+toggle)][||]] /MetaQmailQueueOfQmailInject/ =Enum Values: fullRun, dryRun, runDebug=  [[elisp:(org-cycle)][| ]]
#+end_org """
@enum.unique
class MetaQmailQueueOfQmailInject(enum.Enum):
####+END:
    defaultQmailInject='defaultQmailInject',
    pythonQmailRemote='pythonQmailRemote',
    defaultQmailRemote='defaultQmailRemote',


####+BEGIN: b:py3:cs:func/typing :funcName "handOffToMetaQmailQueue" :funcType "extTyped" :deco "track"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  F-T-extTyped [[elisp:(outline-show-subtree+toggle)][||]] /handOffToMetaQmailQueue/  deco=track  [[elisp:(org-cycle)][| ]]
#+end_org """
@cs.track(fnLoc=True, fnEntry=True, fnExit=True)
def handOffToMetaQmailQueue(
####+END:
        msg,
        metaQmailQueueOfQmailInject:  str,
):
    """ #+begin_org
** [[elisp:(org-cycle)][| *DocStr | ] We are simulating the end handoff stage of qmailInject
    in submission context.
    Should be converted to be an operation
    #+end_org """

    senderAddr = x822Lib.msgSender(msg)
    senderHost = senderAddr.split('@')[1]
    allRecipients = x822Lib.msgAllRecipients(msg)
    #print(allRecipients)

    qmailRemoteArgs = [senderHost, senderAddr,]

    for eachRecipient in allRecipients:
        qmailRemoteArgs.append(eachRecipient[1])


    #b_io.tm.here("qmail Inject \n{msgStr}".format(msgStr=msg.as_string()))
    outcome = b.op.Outcome()

    injectionProgramCmnd = ""
    injectionProgramArgs = []

    if metaQmailQueueOfQmailInject == MetaQmailQueueOfQmailInject.defaultQmailInject.value[0]:
        injectionProgramCmnd =  "qmail-inject-bisos.cs"
        if cs.runArgs.isRunModeDryRun():
            injectionProgramArgs.append('-n')
    elif metaQmailQueueOfQmailInject == MetaQmailQueueOfQmailInject.defaultQmailRemote.value[0]:
        injectionProgramCmnd =  "qmail-remote-bisos.cs"
        injectionProgramArgs =  qmailRemoteArgs
    elif metaQmailQueueOfQmailInject == MetaQmailQueueOfQmailInject.pythonQmailRemote.value[0]:
        qmailRemoteWrapper.qmailRemoteWithMsg(msg, qmailRemoteArgs)
    else:
        print(f"NOTYET {metaQmailQueueOfQmailInject}")
        return(b_io.eh.badOutcome(outcome))

    if injectionProgramCmnd:
        commandLine = injectionProgramCmnd + " " +  " ".join(injectionProgramArgs)
        # print(commandLine)

        if b.subProc.Op(outcome=outcome, log=1).bash(
            f"""{commandLine}""",
            stdin=msg.as_string(),
        ).isProblematic():
            print(outcome.stderr)
            return(b_io.eh.badOutcome(outcome))

        #if outcome.stdout: icm.ANN_note("Stdout: " +  outcome.stdout)
        #if outcome.stderr: icm.ANN_note("Stderr: " +  outcome.stderr)


    return outcome.set(
        opError=b.OpError.Success,
        opResults=None,
    )


####+BEGIN: bx:cs:py3:section :title "CS-Commands"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  /Section/    [[elisp:(outline-show-subtree+toggle)][||]] *CS-Commands*  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:


####+BEGIN: b:py3:cs:cmnd/classHead :cmndName "prepAndQmailInject" :cmndType ""  :comment "Inject as a cmnd" :parsMand "" :parsOpt "" :argsMin 0 :argsMax 9999 :pyInv ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  CmndSvc-   [[elisp:(outline-show-subtree+toggle)][||]] <<prepAndQmailInject>>  *Inject as a cmnd*  =verify= argsMax=9999 ro=cli   [[elisp:(org-cycle)][| ]]
#+end_org """
class prepAndQmailInject(cs.Cmnd):
    cmndParamsMandatory = [ ]
    cmndParamsOptional = [ ]
    cmndArgsLen = {'Min': 0, 'Max': 9999,}

    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmnd(self,
             rtInv: cs.RtInvoker,
             cmndOutcome: b.op.Outcome,
             argsList: typing.Optional[list[str]]=None,  # CsArgs
    ) -> b.op.Outcome:
        """Inject as a cmnd"""
        callParamsDict = {}
        if self.invocationValidate(rtInv, cmndOutcome, callParamsDict, argsList).isProblematic():
            return b_io.eh.badOutcome(cmndOutcome)
        cmndArgsSpecDict = self.cmndArgsSpec()
####+END:
        """ #+begin_org
** [[elisp:(org-cycle)][| *DocStr | ] Interface to qmailInject but through a cmnd
        #+end_org """

        mailInput = b_io.stdin.read()
        qmailInject(mailInput, argsList)

        return(cmndOutcome)

####+BEGIN: b:py3:cs:cmnd/classHead :cmndName "monolithicSend" :cmndType ""  :comment "Inject as a cmnd" :parsMand "" :parsOpt "" :argsMin 0 :argsMax 9999 :pyInv "mailInput"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  CmndSvc-   [[elisp:(outline-show-subtree+toggle)][||]] <<monolithicSend>>  *Inject as a cmnd*  =verify= argsMax=9999 ro=cli pyInv=mailInput   [[elisp:(org-cycle)][| ]]
#+end_org """
class monolithicSend(cs.Cmnd):
    cmndParamsMandatory = [ ]
    cmndParamsOptional = [ ]
    cmndArgsLen = {'Min': 0, 'Max': 9999,}

    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmnd(self,
             rtInv: cs.RtInvoker,
             cmndOutcome: b.op.Outcome,
             argsList: typing.Optional[list[str]]=None,  # CsArgs
             mailInput: typing.Any=None,   # pyInv Argument
    ) -> b.op.Outcome:
        """Inject as a cmnd"""
        callParamsDict = {}
        if self.invocationValidate(rtInv, cmndOutcome, callParamsDict, argsList).isProblematic():
            return b_io.eh.badOutcome(cmndOutcome)
        cmndArgsSpecDict = self.cmndArgsSpec()
####+END:
        """ #+begin_org
** [[elisp:(org-cycle)][| *DocStr | ] Was previously called prepAndQmailRemote
        #+end_org """

        if not mailInput:
            mailInput = b_io.stdin.read()
        msg = qmailInjectPrep(mailInput, argsList)

        handOffToMetaQmailQueue(
            msg,
            MetaQmailQueueOfQmailInject.pythonQmailRemote.value[0],
        )

        return(cmndOutcome)


####+BEGIN: b:py3:cs:func/typing :funcName "qmailInjectPrep" :funcType "ExtTyp" :deco "track"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  F-T-ExtTyp [[elisp:(outline-show-subtree+toggle)][||]] /qmailInjectPrep/  deco=track  [[elisp:(org-cycle)][| ]]
#+end_org """
@cs.track(fnLoc=True, fnEntry=True, fnExit=True)
def qmailInjectPrep(
####+END:
        mailInput: str,
        argsList: list[str],
) -> EmailMessage:
    """#+begin_org
** [[elisp:(org-cycle)][| *DocStr | ] This =qmailInject= is a plugin replacement (a wrapper) for qmail-remote.

    #+end_org """


    body = mailInput

    fd, tmpPath = tempfile.mkstemp(suffix=".mail", prefix="qmail-inject-")
    try:
        with os.fdopen(fd, 'w') as tmp:
            tmp.write(f"argsList={argsList}\n")
            tmp.write(body)
    finally:
        #os.remove(tmpPath)
        #print(f"{tmpPath}")
        pass

    msgParser = Parser()
    msg = msgParser.parsestr(body)

    fromaddr = parseaddr(msg["from"])[1]

    uid = os.getuid()
    pid = os.getpid()

    uidName = pwd.getpwuid(uid)[0]

    aasMarmeeBase = aasMailFps.marmeeBaseForUsageAcct(
        uidName,
    )

    marmee_bpoId, marmee_bpoRunEnv = aasOutMailFps.marmeeAasOutMailAddrFind(
        aasMarmeeBase,
        fromaddr,
    )

    bpoId = msg.get_all("x-bpoid")
    if not bpoId:
        msg['X-bpoId'] = marmee_bpoId
        bpoId = [marmee_bpoId]

    bpoRunEnv = msg.get_all("x-bporunenv")
    if not bpoRunEnv:
        msg['X-bpoRunEnv'] = marmee_bpoRunEnv
        bpoRunEnv = [marmee_bpoRunEnv]

    #print(f"bpoId={bpoId} bpoRunEnv={bpoRunEnv}")

    outMailFps = b.pattern.sameInstance(
        aasOutMailFps.AasOutMail_FPs,
        bpoId=bpoId[0],
        envRelPath=bpoRunEnv[0],
    )

    credsFps = b.pattern.sameInstance(
        gmailOauth2.AasMail_googleCreds_FPs,
        bpoId=bpoId[0],
        envRelPath=bpoRunEnv[0],
    )
    client_id = credsFps.fps_getParam('googleCreds_client_id').parValueGet()
    client_secret = credsFps.fpCrypt_getParam('googleCreds_client_secret').parValueGet().decode("utf-8")
    refresh_token = credsFps.fpCrypt_getParam('googleCreds_refresh_token').parValueGet().decode("utf-8")

    msg['X-Oauth2-Client-Id'] = client_id
    msg['X-Oauth2-Client-Secret'] = client_secret
    msg['X-Oauth2-Refresh-Token'] = refresh_token

    #print(msg)

    return msg


####+BEGIN: b:py3:cs:func/typing :funcName "prepedQmailSubProc" :funcType "ExtTyp" :deco "track"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  F-T-ExtTyp [[elisp:(outline-show-subtree+toggle)][||]] /prepedQmailInvoke/  deco=track  [[elisp:(org-cycle)][| ]]
#+end_org """
@cs.track(fnLoc=True, fnEntry=True, fnExit=True)
def prepedQmailInvoke(
####+END:
        msg: EmailMessage,
        injectionProgram,
        injectionProgramArgs,
) -> None:
    """#+begin_org
** [[elisp:(org-cycle)][| *DocStr | ] This =qmailInject= is a plugin replacement (a wrapper) for qmail-remote.
    #+end_org """

    outcome = x822Out.injectMsgWithQmailVariant(
        msg,
        injectionProgram,
        injectionProgramArgs,
    )

    return outcome

####+BEGIN: b:py3:cs:framework/endOfFile :basedOn "classification"
""" #+begin_org
* [[elisp:(org-cycle)][| *End-Of-Editable-Text* |]] :: emacs and org variables and control parameters
#+end_org """

#+STARTUP: showall

### local variables:
### no-byte-compile: t
### end:
####+END:



# -*- coding: utf-8 -*-

""" #+begin_org
* ~[Summary]~ :: A =CS-Lib=
#+end_org """

####+BEGIN: b:py3:cs:file/dblockControls :classification "cs-u"
""" #+begin_org
* [[elisp:(org-cycle)][| /Control Parameters Of This File/ |]] :: dblk ctrls classifications=cs-u
#+BEGIN_SRC emacs-lisp
(setq-local b:dblockControls t) ; (setq-local b:dblockControls nil)
(put 'b:dblockControls 'py3:cs:Classification "cs-u") ; one of cs-mu, cs-u, cs-lib, b-lib, pyLibPure
#+END_SRC
#+RESULTS:
: cs-mu
#+end_org """
####+END:

####+BEGIN: b:prog:file/proclamations :outLevel 1
""" #+begin_org
* *[[elisp:(org-cycle)][| Proclamations |]]* :: Libre-Halaal Software --- Part Of BISOS ---  Poly-COMEEGA Format.
** This is Libre-Halaal Software. © Neda Communications, Inc. Subject to AGPL.
** It is part of BISOS (ByStar Internet Services OS)
** Best read and edited  with Blee in Poly-COMEEGA (Polymode Colaborative Org-Mode Enhance Emacs Generalized Authorship)
#+end_org """
####+END:

####+BEGIN: b:prog:file/particulars :authors ("./inserts/authors-mb.org")
""" #+begin_org
* *[[elisp:(org-cycle)][| Particulars |]]* :: Authors, version
** This File: /bisos/git/auth/bxRepos/bisos-pip/marmee/py3/bisos/marmee/aasInMailOfflineimap.py
** Authors: Mohsen BANAN, http://mohsen.banan.1.byname.net/contact
#+end_org """
####+END:

####+BEGIN: b:python:file/particulars-csInfo :status "inUse"
""" #+begin_org
* *[[elisp:(org-cycle)][| Particulars-csInfo |]]*
#+end_org """
import typing
csInfo: typing.Dict[str, typing.Any] = { 'moduleName': ['marmeeQmail'], }
csInfo['version'] = '202212053620'
csInfo['status']  = 'inUse'
csInfo['panel'] = 'marmeeQmail-Panel.org'
csInfo['groupingType'] = 'IcmGroupingType-pkged'
csInfo['cmndParts'] = 'IcmCmndParts[common] IcmCmndParts[param]'
####+END:

csInfo['description'] = """ #+begin_org
* /[[elisp:(org-cycle)][| Description |]]/ :: [[file:/bisos/git/auth/bxRepos/blee-binders/bisos-core/COMEEGA/_nodeBase_/fullUsagePanel-en.org][BISOS COMEEGA Panel]]
Module description comes here.
** Relevant Panels:
** Status: In use with blee3
** /[[elisp:(org-cycle)][| Planned Improvements |]]/ :
*** TODO complete fileName in particulars.
#+end_org """

####+BEGIN: b:prog:file/orgTopControls :outLevel 1
""" #+begin_org
* [[elisp:(org-cycle)][| Controls |]] :: [[elisp:(delete-other-windows)][(1)]] | [[elisp:(show-all)][Show-All]]  [[elisp:(org-shifttab)][Overview]]  [[elisp:(progn (org-shifttab) (org-content))][Content]] | [[elisp:(blee:ppmm:org-mode-toggle)][Nat]] | [[elisp:(bx:org:run-me)][Run]] | [[elisp:(bx:org:run-me-eml)][RunEml]] | [[elisp:(progn (save-buffer) (kill-buffer))][S&Q]]  [[elisp:(save-buffer)][Save]]  [[elisp:(kill-buffer)][Quit]] [[elisp:(org-cycle)][| ]]
** /Version Control/ ::  [[elisp:(call-interactively (quote cvs-update))][cvs-update]]  [[elisp:(vc-update)][vc-update]] | [[elisp:(bx:org:agenda:this-file-otherWin)][Agenda-List]]  [[elisp:(bx:org:todo:this-file-otherWin)][ToDo-List]]
#+end_org """
####+END:

####+BEGIN: b:python:file/workbench :outLevel 1
""" #+begin_org
* [[elisp:(org-cycle)][| Workbench |]] :: [[elisp:(python-check (format "/bisos/venv/py3/bisos3/bin/python -m pyclbr %s" (bx:buf-fname))))][pyclbr]] || [[elisp:(python-check (format "/bisos/venv/py3/bisos3/bin/python -m pydoc ./%s" (bx:buf-fname))))][pydoc]] || [[elisp:(python-check (format "/bisos/pipx/bin/pyflakes %s" (bx:buf-fname)))][pyflakes]] | [[elisp:(python-check (format "/bisos/pipx/bin/pychecker %s" (bx:buf-fname))))][pychecker (executes)]] | [[elisp:(python-check (format "/bisos/pipx/bin/pycodestyle %s" (bx:buf-fname))))][pycodestyle]] | [[elisp:(python-check (format "/bisos/pipx/bin/flake8 %s" (bx:buf-fname))))][flake8]] | [[elisp:(python-check (format "/bisos/pipx/bin/pylint %s" (bx:buf-fname))))][pylint]]  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:


####+BEGIN: b:py3:cs:orgItem/basic :type "=PyImports= " :title "*Py Library IMPORTS*" :comment "-- with classification based framework/imports"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  =PyImports=  [[elisp:(outline-show-subtree+toggle)][||]] *Py Library IMPORTS* -- with classification based framework/imports  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:

####+BEGIN: b:py3:cs:framework/imports :basedOn "classification"
""" #+begin_org
** Imports Based On Classification=cs-lib
#+end_org """
from bisos import b
from bisos.b import cs
from bisos.b import b_io

####+END:

import collections

import collections
import pathlib
import os
#import shutil

from bisos.common import csParam

# from bisos.qmail import maildrop

from bisos.basics import pyRunAs
from bisos.basics import basics
# from bisos.common import lines

from bisos.qmail import qmailControl

import enum
import pathlib

import pwd


####+BEGIN: bx:dblock:python:section :title "Enumerations"
"""
*  [[elisp:(beginning-of-buffer)][Top]] ############## [[elisp:(blee:ppmm:org-mode-toggle)][Nat]] [[elisp:(delete-other-windows)][(1)]]    *Enumerations*  [[elisp:(org-cycle)][| ]]  [[elisp:(org-show-subtree)][|=]]
"""
####+END:

####+BEGIN: bx:dblock:python:enum :enumName "qmailCmndExit_Code" :comment ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Enum       [[elisp:(outline-show-subtree+toggle)][||]] /qmailCmndExit_Code/  [[elisp:(org-cycle)][| ]]
#+end_org """
@enum.unique
class qmailCmndExit_Code(enum.Enum):
####+END:
    qmailCmndExitSuccess=0
    qmailCmndExitSuccessFinish=99
    qmailCmndExitFailHard=100
    qmailCmndExitFailSoft=111

####+BEGIN: bx:dblock:python:enum :enumName "acctAddr_Type" :comment ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Enum       [[elisp:(outline-show-subtree+toggle)][||]] /acctAddr_Type/  [[elisp:(org-cycle)][| ]]
#+end_org """
@enum.unique
class acctAddr_Type(enum.Enum):
####+END:
    finalDelivery = 'finalDelivery'



####+BEGIN: bx:cs:py3:section :title "Public Classes"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  /Section/    [[elisp:(outline-show-subtree+toggle)][||]] *Public Functions*  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:


####+BEGIN: b:py3:class/decl :className "QmailInstallationSingleton" :superClass "object" :comment "Abstraction of a  Interface" :classType "basic"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Cls-basic  [[elisp:(outline-show-subtree+toggle)][||]] /QmailInstallationSingleton/  superClass=object =Abstraction of a  Interface=  [[elisp:(org-cycle)][| ]]
#+end_org """
class QmailInstallationSingleton(object):
####+END:
    """
** Abstraction of
"""
    _instance = None

    # Singleton using New
    def __new__(cls):
        if cls._instance is None:
            # print('Creating the object')
            cls._instance = super(QmailInstallationSingleton, cls).__new__(cls)
            # Put any initialization here.
        return cls._instance

    def __init__(
            self,
    ):
        self._varDir = pathlib.Path("MISSING")
        varDir =  pathlib.Path("/var/qmail")
        if varDir.is_dir():
            self._varDir = varDir
        else:
            varDir =  pathlib.Path("/var/lib/qmail")
            if varDir.is_dir():
                self._varDir = varDir

        self._binBaseDir = pathlib.Path("MISSING")
        binBaseDir =  pathlib.Path("/var/qmail/bin")
        if binBaseDir.is_dir():
            self._binBaseDir = binBaseDir
        else:
            if pathlib.Path("/usr/sbin/qmail-inject").is_file():
                self._binBaseDir = pathlib.Path("/usr/sbin")

        self._controlBaseDir = pathlib.Path("MISSING")
        controlBaseDir =  pathlib.Path("/var/qmail/control")
        if controlBaseDir.is_dir():
            self._controlBaseDir = controlBaseDir
        else:
            controlBaseDir =  pathlib.Path("/etc/qmail")
            if controlBaseDir.is_dir():
                self._controlBaseDir = controlBaseDir

        self._usersBaseDir = self._varDir.joinpath("users")

    def fullProgPath(self, progName: str,):
        return str(self._binBaseDir.joinpath(progName))

    @property
    def varDir(self) -> pathlib.Path:
        return self._varDir

    @property
    def binBaseDir(self) -> pathlib.Path:
        return self._binBaseDir

    @property
    def controlBaseDir(self) -> pathlib.Path:
        return self._controlBaseDir

    @property
    def usersBaseDir(self) -> pathlib.Path:
        return self._usersBaseDir


installation = QmailInstallationSingleton()


####+BEGIN: b:py3:class/decl :className "LocalDeliveryAcct" :superClass "object" :comment "Abstraction of a qmail delivery account a set of statics" :classType "basic"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Cls-basic  [[elisp:(outline-show-subtree+toggle)][||]] /LocalDeliveryAcct/  superClass=object =Abstraction of a qmail delivery account a set of statics=  [[elisp:(org-cycle)][| ]]
#+end_org """
class LocalDeliveryAcct(object):
####+END:
    """
** Abstraction of a Local Delivery Acct
"""

    def __init__(
            self,
    ):
        pass

####+BEGIN: b:py3:cs:method/typing :methodName "ensureUsersBaseDir" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /ensureUsersBaseDir/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def ensureUsersBaseDir(
####+END:
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:

        if inOutcome is None:
            outcome = b.op.Outcome()
        else:
            outcome = inOutcome

        result: typing.Any = None

        if installation.usersBaseDir.is_dir():
            result = f"""usersBaseDir={str(installation.usersBaseDir)} exists, creation skipped"""
            return outcome.set(opResults=result)

        pyRunAs.as_root_osSystem(
            f"""mkdir {str(installation.usersBaseDir)}"""
        )
        pyRunAs.as_root_osSystem(
            f"""chmod 755 {str(installation.usersBaseDir)}"""
        )

        includePath = installation.usersBaseDir.joinpath("include")
        if not includePath.is_file():
            basics.File.writeWithStr(includePath, "", perhapsAsRoot=True)

        assignPath = installation.usersBaseDir.joinpath("assign")
        if not assignPath.is_file():
            basics.File.writeWithStr(assignPath, "", perhapsAsRoot=True)

        result = f"""Created usersBaseDir={str(installation.usersBaseDir)}"""
        return outcome.set(opResults=result)


####+BEGIN: b:py3:cs:method/typing :methodName "newUserProc" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /newUserProc/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def newUserProc(
####+END:
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:

        results: str = ""

        if inOutcome is None:
            outcome = b.op.Outcome()
            if outcome.results:
                results = outcome.results
        else:
            outcome = inOutcome

        qmailPw2uProgram = installation.binBaseDir.joinpath("qmail-pw2u")
        qmailNewuProgram = installation.binBaseDir.joinpath("qmail-newu")

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
cat /etc/passwd | {qmailPw2uProgram}  > {installation.usersBaseDir}/assign
{qmailNewuProgram}
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))

        results = results + "LocalDeliveryAcct.newUserProc: processed users\n"
        return outcome.set(opResults=results)

####+BEGIN: b:py3:cs:method/typing :methodName "add" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /add/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def add(
####+END:
            user: str,
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:

        results: str = ""

        if inOutcome is None:
            outcome = b.op.Outcome()
            if outcome.results:
                results = outcome.results
        else:
            outcome = inOutcome

        try:
            pwd.getpwnam(user)
        except KeyError:
            outcome.error = 1
            outcome.errInfo = "Missing Passwd Account"
            return outcome

        LocalDeliveryAcct.ensureUsersBaseDir()

        includePath = installation.usersBaseDir.joinpath("include")
        if basics.FileLine.isIn(includePath, user):
            results = f"""basics.FileLine.isIn: {user} already in {includePath}, Addition skipped"""
            return outcome.set(opResults=results)

        basics.FileLine.appendIfNotThere(includePath, user, perhapsAsRoot=True)

        LocalDeliveryAcct.newUserProc(outcome)

        results = outcome.results + f"""LocalDeliveryAcct.add: Added {user} to {includePath}"""
        return outcome.set(opResults=results)

####+BEGIN: b:py3:cs:method/typing :methodName "delete" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /delete/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def delete(
####+END:
            user: str,
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:

        results: str = ""

        if inOutcome is None:
            outcome = b.op.Outcome()
            if outcome.results:
                results = outcome.results
        else:
            outcome = inOutcome

        try:
            pwd.getpwnam(user)
        except KeyError:
            outcome.error = 1
            outcome.errInfo = "Missing Passwd Account"
            return outcome

        LocalDeliveryAcct.ensureUsersBaseDir()

        includePath = installation.usersBaseDir.joinpath("include")
        if not basics.FileLine.isIn(includePath, user):
            results = f"""basics.FileLine.isIn: {user} not in {includePath}, deletion skipped"""
            return outcome.set(opResults=results)

        basics.FileLine.remove(includePath, user, perhapsAsRoot=True)

        LocalDeliveryAcct.newUserProc(outcome)

        results = outcome.results + f"""LocalDeliveryAcct.add: removed {user} from {includePath}"""
        return outcome.set(opResults=results)

####+BEGIN: b:py3:cs:method/typing :methodName "verify" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /verify/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def verify(
####+END:
            user: str,
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:

        results: str = ""

        if inOutcome is None:
            outcome = b.op.Outcome()
            if outcome.results:
                results = outcome.results
        else:
            outcome = inOutcome

        try:
            pwd.getpwnam(user)
        except KeyError:
            outcome.error = 1
            outcome.errInfo = "Missing Passwd Account"
            return outcome

        assignPath = installation.usersBaseDir.joinpath("assign")
        if basics.FileLine.isIn(assignPath, f":{user}:"):
            results = f"""LocalDeliveryAcct.verify: {user} is in {assignPath}"""
            return outcome.set(opResults=results)
        else:
            outcome.error = 2
            outcome.errInfo = "Not in assign file"
            return outcome

####+BEGIN: b:py3:cs:method/typing :methodName "mainDomainGet" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /mainDomainGet/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def mainDomainGet(
####+END:
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:

        results: str = ""

        if inOutcome is None:
            outcome = b.op.Outcome()
            if outcome.results:
                results = outcome.results
        else:
            outcome = inOutcome

        results = qmailControl.QCFV_QmailSend().locals
        return outcome.set(opResults=results)


####+BEGIN: b:py3:class/decl :className "LocalDeliveryAction" :superClass "object" :comment "Invoke qmail-local with static methods" :classType "basic"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Cls-basic  [[elisp:(outline-show-subtree+toggle)][||]] /LocalDeliveryAction/  superClass=object =Invoke qmail-local with static methods=  [[elisp:(org-cycle)][| ]]
#+end_org """
class LocalDeliveryAction(object):
####+END:
    """
** Abstraction of a Local Delivery Acct
"""

    def __init__(
            self,
    ):
        pass

####+BEGIN: b:py3:cs:method/typing :methodName "testShow" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /testShow/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def testShow(
####+END:
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:
        """
  mmaQmail_dotQmailFile=${opAcct_homeDir}/.qmail
  typeset dotQmailFile=${mmaQmail_dotQmailFile}

  EH_assert [[ -f ${dotQmailFile}-${localPart} ]]

  opDo eval "qmail-local -n ${opAcct_name} ${opAcct_homeDir} ${opAcct_name} '-' '${localPart}' '' '' ${dotQmailFile}"
        """

        if inOutcome is None:
            outcome = b.op.Outcome()
        else:
            outcome = inOutcome

        result: typing.Any = None
        return outcome.set(opResults=result)


####+BEGIN: b:py3:class/decl :className "Maildir" :superClass "object" :comment "Abstraction of Maildir CRUD" :classType "basic"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Cls-basic  [[elisp:(outline-show-subtree+toggle)][||]] /Maildir/  superClass=object =Abstraction of Maildir CRUD=  [[elisp:(org-cycle)][| ]]
#+end_org """
class Maildir(object):
####+END:
    """
** Abstraction of Maildir CRUD
"""

    def __init__(
            self,
    ):
        pass

####+BEGIN: b:py3:cs:method/typing :methodName "create" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /create/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def create(
####+END:
            qAddrAcct: str,
            maildirsList: list[str],
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:
        """
        maildirFullPath = pathlib.Path(os.path.join(dataBase, 'maildir'))

        if not maildirFullPath.is_dir():
            maildirFullPath.mkdir(parents=True, exist_ok=True)

        for eachMaildir in maildirNames:
            hereMaildirMainBase = maildirFullPath.joinpath(eachMaildir)

            if not hereMaildirMainBase.joinpath('new').is_dir():
                if not (resStr := b.subProc.WOpW(invedBy=self, log=1).bash(
                    f'maildirmake {hereMaildirMainBase}',
                ).stdout):
                    b_io.eh.badOutcome(cmndOutcome)
                    #return(b_io.eh.badOutcome(cmndOutcome))
        """

        if inOutcome is None: outcome = b.op.Outcome()
        else: outcome = inOutcome

        result: typing.Any = None

        for eachMaildir in maildirsList:
            hereMaildirMainBase = pathlib.Path(eachMaildir)

            if not hereMaildirMainBase.joinpath('new').is_dir():
                if not (resStr := b.subProc.WOpW(invedBy=None, log=1).bash(
                    f'maildirmake {hereMaildirMainBase}',
                ).stdout):
                    b_io.eh.badOutcome(outcome)

        return outcome.set(opResults=result)

####+BEGIN: b:py3:cs:method/typing :methodName "verify" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /verify/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def verify(
####+END:
            qAddrAcct: str,
            maildirsList: list[str],
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:
        """
        """

        return "NOTYET"

        if inOutcome is None: outcome = b.op.Outcome()
        else: outcome = inOutcome

        result: typing.Any = None

        for eachMaildir in maildirsList:
            hereMaildirMainBase = pathlib.Path(eachMaildir)

            if not hereMaildirMainBase.joinpath('new').is_dir():
                if not (resStr := b.subProc.WOpW(invedBy=None, log=1).bash(
                    f'maildirmake {hereMaildirMainBase}',
                ).stdout):
                    b_io.eh.badOutcome(outcome)

        return outcome.set(opResults=result)

####+BEGIN: b:py3:cs:method/typing :methodName "delete" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /delete/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def delete(
####+END:
            qAddrAcct: str,
            maildirsList: list[str],
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:
        """
        """

        return "NOTYET"
        
        if inOutcome is None: outcome = b.op.Outcome()
        else: outcome = inOutcome

        result: typing.Any = None

        for eachMaildir in maildirsList:
            hereMaildirMainBase = pathlib.Path(eachMaildir)

            if not hereMaildirMainBase.joinpath('new').is_dir():
                if not (resStr := b.subProc.WOpW(invedBy=None, log=1).bash(
                    f'maildirmake {hereMaildirMainBase}',
                ).stdout):
                    b_io.eh.badOutcome(outcome)

        return outcome.set(opResults=result)


####+BEGIN: b:py3:class/decl :className "DotQmailFile" :superClass "object" :comment "Abstraction of a dotQmail" :classType "basic"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Cls-basic  [[elisp:(outline-show-subtree+toggle)][||]] /DotQmailFile/  superClass=object =Abstraction of a dotQmail=  [[elisp:(org-cycle)][| ]]
#+end_org """
class DotQmailFile(object):
####+END:
    """
** Abstraction of a dotQmail
"""

    def __init__(
            self,
            qAddrAcct,
            localPart,
    ):

        self.qAddrAcct = qAddrAcct
        self.localPart = localPart

####+BEGIN: b:py3:cs:method/typing :methodName "acctPath" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /acctPath/ deco=default  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def acctPath(
####+END:
            self,
    ) -> pathlib.Path:
        return pathlib.Path(os.path.expanduser(f'~{self.qAddrAcct}'))

####+BEGIN: b:py3:cs:method/typing :methodName "nameOfLocalPart" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /nameOfLocalPart/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def nameOfLocalPart(
####+END:
            self,
    ) -> pathlib.Path:
        if self.localPart =="":
            return pathlib.Path(f".qmail")
        else:
            return pathlib.Path(f".qmail-{self.localPart}")

####+BEGIN: b:py3:cs:method/typing :methodName "filePath" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /filePath/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def filePath(
####+END:
            self,
    ) -> pathlib.Path:
        return (
            self.acctPath().joinpath(
               self.nameOfLocalPart()
            )
        )

####+BEGIN: b:py3:cs:method/typing :methodName "contentRead" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /contentRead/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def contentRead(
####+END:
            self,
    ) -> str:
        return (
            pyRunAs.as_root_readFromFile(
                self.filePath()
            )
        )

####+BEGIN: b:py3:cs:method/typing :methodName "contentWrite" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /contentWrite/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def contentWrite(
####+END:
            self,
            content: str,
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:
        """
        """
        if inOutcome is None: outcome = b.op.Outcome()
        else: outcome = inOutcome.adopt()

        result: typing.Any = None

        pyRunAs.as_root_writeToFile(
            self.filePath(),
            content,
        )
        pyRunAs.as_root_osSystem(
            f"""chown {self.qAddrAcct} {self.filePath()}"""
        )
        pyRunAs.as_root_osSystem(
            f"""chmod 600 {self.filePath()}"""
        )

        outcome.report(f"Wrote dotQmail={self.filePath()}")

        os.system(f"ls -l {self.filePath()}")

        outcome.set(opResults=result)


####+BEGIN: b:py3:cs:method/typing :methodName "addMaildirLine" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /addMaildirLine/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def addMaildirLine(
####+END:
            self,
            maildirPath: str,
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:
        """
        """
        if inOutcome is None: outcome = b.op.Outcome()
        else: outcome = inOutcome.adopt()

        result: typing.Any = None

        if not self.filePath().is_file():
            basics.File.writeWithStr(self.filePath(), "", perhapsAsRoot=True)
        else:
            if basics.FileLine.isIn(self.filePath(), maildirPath, perhapsAsRoot=True):
                outcome.report(f"""basics.FileLine.isIn: {maildirPath} already in {self.filePath()}, Addition skipped""")
                return outcome.set(opResults=result)

        basics.FileLine.appendIfNotThere(self.filePath(), maildirPath, perhapsAsRoot=True)

        pyRunAs.as_root_osSystem(
            f"""chown {self.qAddrAcct} {self.filePath()}"""
        )
        pyRunAs.as_root_osSystem(
            f"""chmod 600 {self.filePath()}"""
        )
        outcome.report(f"""Added {maildirPath} to {self.filePath()}""")
        return outcome.set(opResults=result)

####+BEGIN: b:py3:cs:method/typing :methodName "addForwardLine" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /addForwardLine/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def addForwardLine(
####+END:
            self,
            destAddr: str,
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:
        """
        """
        if inOutcome is None: outcome = b.op.Outcome()
        else: outcome = inOutcome.adopt()

        result: typing.Any = None

        if not self.filePath().is_file():
            basics.File.writeWithStr(self.filePath(), "", perhapsAsRoot=True)
        else:
            if basics.FileLine.isIn(self.filePath(), destAddr, perhapsAsRoot=True):
                outcome.report(f"""basics.FileLine.isIn: {destAddr} already in {self.filePath()}, Addition skipped""")
                return outcome.set(opResults=result)

        basics.FileLine.appendIfNotThere(self.filePath(), f"&{destAddr}", perhapsAsRoot=True)

        pyRunAs.as_root_osSystem(
            f"""chown {self.qAddrAcct} {self.filePath()}"""
        )
        pyRunAs.as_root_osSystem(
            f"""chmod 600 {self.filePath()}"""
        )
        outcome.report(f"""Added {destAddr} to {self.filePath()}""")
        return outcome.set(opResults=result)


####+BEGIN: b:py3:cs:method/typing :methodName "deleteLine" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /deleteLine/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def deleteLine(
####+END:
           self,
            line: str,
            inOutcome: b.op.Outcome | None =None,
    )-> b.op.Outcome:
        """
        """
        if inOutcome is None: outcome = b.op.Outcome()
        else: outcome = inOutcome.adopt()

        result: typing.Any = None


        result = basics.FileLine.remove(self.filePath(), line, perhapsAsRoot=True)

        pyRunAs.as_root_osSystem(
            f"""chown {self.qAddrAcct} {self.filePath()}"""
        )
        pyRunAs.as_root_osSystem(
            f"""chmod 600 {self.filePath()}"""
        )
        if result == True:
            outcome.report(f"""Removed  {line} from {self.filePath()}""")
        else:
            outcome.report(f"""{self.filePath()} does not contain {line}, removal skipped""")

        return outcome.set(opResults=result)



####+BEGIN: b:py3:cs:method/typing :methodName "deliveryStatus" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /deliveryStatus/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def deliveryStatus(
####+END:
            self,
    ) -> None:
        """
  findResult=$( find ${opAcct_homeDir} -maxdepth 0 -perm -01000 )
  if [[ ${findResult} == ${opAcct_homeDir} ]] ; then
    ANT_raw "Sticky Bit Set, Delivery is On Hold"
    opDo ls -ld ${opAcct_homeDir}
  else
    ANT_raw "Sticky Bit NOT Set, Normal Delivery"
    opDo ls -ld ${opAcct_homeDir}
  fi
         """
        pass

####+BEGIN: b:py3:cs:method/typing :methodName "deliveryPause" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /deliveryPause/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def deliveryPause(
####+END:
            self,
    ) -> None:
        """
          opDo chmod +t ${opAcct_homeDir}
  opDo ls -ld ${opAcct_homeDir}
         """
        pass

####+BEGIN: b:py3:cs:method/typing :methodName "deliveryResume" :deco "default"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /deliveryResume/  deco=default  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def deliveryResume(
####+END:
            self,
    ) -> None:
        """
          opDo chmod -t ${opAcct_homeDir}
  opDo ls -ld ${opAcct_homeDir}
         """
        pass


####+BEGIN: b:py3:class/decl :className "VirDomEntry" :superClass "object" :comment "Abstraction of a VidrDom Entry" :classType "basic"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Cls-basic  [[elisp:(outline-show-subtree+toggle)][||]] /VirDomEntry/  superClass=object =Abstraction of a VidrDom Entry=  [[elisp:(org-cycle)][| ]]
#+end_org """
class VirDomEntry(object):
####+END:
    """
** Abstraction of a Local Delivery Acct
"""

    def __init__(
            self,
    ):
        pass

####+BEGIN: b:py3:cs:method/typing :methodName "add" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /add/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def add(
####+END:
            virDomName: str,
            virDomAcct: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
  if USER_isInPasswdFile ${2} ; then
    FN_lineAddToFile "^${1}:${2}\$" "${1}:${2}" ${qmailVirDomFile}
  else
    print "$0: Skipping ${2} because it is not in the passwd file"
    return 2
  fi

  #grep ${2} ${qmailVirDomFile}

  print "pkill -HUP 'qmail-send'"
  pkill -HUP 'qmail-send'
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome

####+BEGIN: b:py3:cs:method/typing :methodName "delete" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /delete/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def delete(
####+END:
            virDomName: str,
            virDomAcct: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
  if USER_isInPasswdFile ${2} ; then
    FN_lineRemoveFromFile "^${1}:${2}\$" ${qmailVirDomFile}
  else
    print "$0: Skipping ${2} because it is not in the passwd file"
    return 2
  fi

  #grep ${2} ${qmailVirDomFile}

  print "pkill -HUP 'qmail-send'"
  pkill -HUP 'qmail-send'
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome


####+BEGIN: b:py3:cs:method/typing :methodName "acctGet" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /acctGet/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def acctGet(
####+END:
            user: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
  typeset acctName=`egrep "^$1:" ${qmailVirDomFile} | cut -d : -f 2`

  if [[ "${acctName}_" == "_" ]] ; then
    EH_problem "$1 not found in ${qmailVirDomFile}"
    return 1
  fi

  print ${acctName}
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome

####+BEGIN: b:py3:cs:method/typing :methodName "domainGet" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /domainGet/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def domainGet(
####+END:
            user: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
   typeset domainName=`egrep ":$1$" ${qmailVirDomFile} | cut -d : -f 1`

  if [[ "${domainName}_" == "_" ]] ; then
    EH_problem "$1 not found in ${qmailVirDomFile}"
    return 1
  fi

  print ${domainName}
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome



####+BEGIN: b:py3:class/decl :className "RcpthostsEntry" :superClass "object" :comment "Abstraction of RcpthostsEntry" :classType "basic"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Cls-basic  [[elisp:(outline-show-subtree+toggle)][||]] /RcpthostsEntry/  superClass=object =Abstraction of RcpthostsEntry=  [[elisp:(org-cycle)][| ]]
#+end_org """
class RcpthostsEntry(object):
####+END:
    """
** Abstraction of RcpthostsEntry
"""

    def __init__(
            self,
    ):
        pass

####+BEGIN: b:py3:cs:method/typing :methodName "add" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /add/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def add(
####+END:
            user: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
  FN_lineAddToFile "^${1}\$" "${1}"  ${qmailControlBaseDir}/rcpthosts

""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome

####+BEGIN: b:py3:cs:method/typing :methodName "delete" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /delete/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def delete(
####+END:
            user: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
  FN_lineRemoveFromFile "^${1}\$" ${qmailControlBaseDir}/rcpthosts
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome

####+BEGIN: b:py3:cs:method/typing :methodName "verify" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /verify/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def verify(
####+END:
            user: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
echo NOTYET
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome

####+BEGIN: b:py3:class/decl :className "VirDom" :superClass "object" :comment "Abstraction of a Virtual Domain" :classType "basic"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Cls-basic  [[elisp:(outline-show-subtree+toggle)][||]] /VirDom/  superClass=object =Abstraction of a Virtual Domain=  [[elisp:(org-cycle)][| ]]
#+end_org """
class VirDom(object):
####+END:
    """
** Abstraction of a Virtual Domain
"""

    def __init__(
            self,
    ):
        pass

####+BEGIN: b:py3:cs:method/typing :methodName "update" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /update/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def update(
####+END:
            virDomName: str,
            virDomAcct: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
  opDoRet mmaQmailLocDeliveryAcctAdd ${2}
  opDoRet mmaQmailVirDomEntryAdd ${1} ${2}
  opDoRet mmaQmailRcpthostsEntryAdd ${1}
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome

####+BEGIN: b:py3:cs:method/typing :methodName "delete" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /delete/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def delete(
####+END:
            virDomName: str,
            virDomAcct: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
  opDoRet mmaQmailLocDeliveryAcctDelete ${2}
  opDoRet mmaQmailVirDomEntryDelete ${1} ${2}
  opDoRet mmaQmailRcpthostsEntryDelete ${1}
 """,
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome


####+BEGIN: b:py3:cs:method/typing :methodName "acctGet" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /acctGet/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def acctGet(
####+END:
            user: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
  typeset acctName=`egrep "^$1:" ${qmailVirDomFile} | cut -d : -f 2`

  if [[ "${acctName}_" == "_" ]] ; then
    EH_problem "$1 not found in ${qmailVirDomFile}"
    return 1
  fi

  print ${acctName}
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome

####+BEGIN: b:py3:cs:method/typing :methodName "domainGet" :deco "staticmethod"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-     [[elisp:(outline-show-subtree+toggle)][||]] /domainGet/  deco=staticmethod  [[elisp:(org-cycle)][| ]]
    #+end_org """
    @staticmethod
    def domainGet(
####+END:
            user: str,
    )-> b.op.Outcome:
        outcome = b.op.Outcome()

        if b.subProc.Op(outcome=outcome, uid="root", log=1).bash(
                f"""\
   typeset domainName=`egrep ":$1$" ${qmailVirDomFile} | cut -d : -f 1`

  if [[ "${domainName}_" == "_" ]] ; then
    EH_problem "$1 not found in ${qmailVirDomFile}"
    return 1
  fi

  print ${domainName}
""",
        ).isProblematic():  return(b_io.eh.badOutcome(outcome))
        return outcome



####+BEGIN: b:py3:cs:framework/endOfFile :basedOn "classification"
""" #+begin_org
* [[elisp:(org-cycle)][| *End-Of-Editable-Text* |]] :: emacs and org variables and control parameters
#+end_org """

#+STARTUP: showall

### local variables:
### no-byte-compile: t
### end:
####+END:



# -*- coding: utf-8 -*-

""" #+begin_org
* ~[Summary]~ :: A =CS-Lib=
#+end_org """

####+BEGIN: b:py3:cs:file/dblockControls :classification "cs-u"
""" #+begin_org
* [[elisp:(org-cycle)][| /Control Parameters Of This File/ |]] :: dblk ctrls classifications=cs-u
#+BEGIN_SRC emacs-lisp
(setq-local b:dblockControls t) ; (setq-local b:dblockControls nil)
(put 'b:dblockControls 'py3:cs:Classification "cs-u") ; one of cs-mu, cs-u, cs-lib, b-lib, pyLibPure
#+END_SRC
#+RESULTS:
: cs-mu
#+end_org """
####+END:

####+BEGIN: b:prog:file/proclamations :outLevel 1
""" #+begin_org
* *[[elisp:(org-cycle)][| Proclamations |]]* :: Libre-Halaal Software --- Part Of BISOS ---  Poly-COMEEGA Format.
** This is Libre-Halaal Software. © Neda Communications, Inc. Subject to AGPL.
** It is part of BISOS (ByStar Internet Services OS)
** Best read and edited  with Blee in Poly-COMEEGA (Polymode Colaborative Org-Mode Enhance Emacs Generalized Authorship)
#+end_org """
####+END:

####+BEGIN: b:prog:file/particulars :authors ("./inserts/authors-mb.org")
""" #+begin_org
* *[[elisp:(org-cycle)][| Particulars |]]* :: Authors, version
** This File: /bisos/git/auth/bxRepos/bisos-pip/marmee/py3/bisos/marmee/aasInMailOfflineimap.py
** Authors: Mohsen BANAN, http://mohsen.banan.1.byname.net/contact
#+end_org """
####+END:

####+BEGIN: b:python:file/particulars-csInfo :status "inUse"
""" #+begin_org
* *[[elisp:(org-cycle)][| Particulars-csInfo |]]*
#+end_org """
import typing
csInfo: typing.Dict[str, typing.Any] = { 'moduleName': ['marmeeQmail'], }
csInfo['version'] = '202212053620'
csInfo['status']  = 'inUse'
csInfo['panel'] = 'marmeeQmail-Panel.org'
csInfo['groupingType'] = 'IcmGroupingType-pkged'
csInfo['cmndParts'] = 'IcmCmndParts[common] IcmCmndParts[param]'
####+END:

csInfo['description'] = """ #+begin_org
* /[[elisp:(org-cycle)][| Description |]]/ :: [[file:/bisos/git/auth/bxRepos/blee-binders/bisos-core/COMEEGA/_nodeBase_/fullUsagePanel-en.org][BISOS COMEEGA Panel]]
Module description comes here.
** Relevant Panels:
** Status: In use with blee3
** /[[elisp:(org-cycle)][| Planned Improvements |]]/ :
*** TODO complete fileName in particulars.
#+end_org """

####+BEGIN: b:prog:file/orgTopControls :outLevel 1
""" #+begin_org
* [[elisp:(org-cycle)][| Controls |]] :: [[elisp:(delete-other-windows)][(1)]] | [[elisp:(show-all)][Show-All]]  [[elisp:(org-shifttab)][Overview]]  [[elisp:(progn (org-shifttab) (org-content))][Content]] | [[elisp:(blee:ppmm:org-mode-toggle)][Nat]] | [[elisp:(bx:org:run-me)][Run]] | [[elisp:(bx:org:run-me-eml)][RunEml]] | [[elisp:(progn (save-buffer) (kill-buffer))][S&Q]]  [[elisp:(save-buffer)][Save]]  [[elisp:(kill-buffer)][Quit]] [[elisp:(org-cycle)][| ]]
** /Version Control/ ::  [[elisp:(call-interactively (quote cvs-update))][cvs-update]]  [[elisp:(vc-update)][vc-update]] | [[elisp:(bx:org:agenda:this-file-otherWin)][Agenda-List]]  [[elisp:(bx:org:todo:this-file-otherWin)][ToDo-List]]
#+end_org """
####+END:

####+BEGIN: b:python:file/workbench :outLevel 1
""" #+begin_org
* [[elisp:(org-cycle)][| Workbench |]] :: [[elisp:(python-check (format "/bisos/venv/py3/bisos3/bin/python -m pyclbr %s" (bx:buf-fname))))][pyclbr]] || [[elisp:(python-check (format "/bisos/venv/py3/bisos3/bin/python -m pydoc ./%s" (bx:buf-fname))))][pydoc]] || [[elisp:(python-check (format "/bisos/pipx/bin/pyflakes %s" (bx:buf-fname)))][pyflakes]] | [[elisp:(python-check (format "/bisos/pipx/bin/pychecker %s" (bx:buf-fname))))][pychecker (executes)]] | [[elisp:(python-check (format "/bisos/pipx/bin/pycodestyle %s" (bx:buf-fname))))][pycodestyle]] | [[elisp:(python-check (format "/bisos/pipx/bin/flake8 %s" (bx:buf-fname))))][flake8]] | [[elisp:(python-check (format "/bisos/pipx/bin/pylint %s" (bx:buf-fname))))][pylint]]  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:


####+BEGIN: b:py3:cs:orgItem/basic :type "=PyImports= " :title "*Py Library IMPORTS*" :comment "-- with classification based framework/imports"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  =PyImports=  [[elisp:(outline-show-subtree+toggle)][||]] *Py Library IMPORTS* -- with classification based framework/imports  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:

####+BEGIN: b:py3:cs:framework/imports :basedOn "classification"
""" #+begin_org
** Imports Based On Classification=cs-lib
#+end_org """
from bisos import b
from bisos.b import cs
from bisos.b import b_io

####+END:

import collections

import collections
import pathlib
import os
#import shutil

#from bisos.bpo import bpoRunBases
#from bisos.bpo import bpo

from bisos.common import csParam

#from bisos.marmee import aasInMailControl
#from bisos.marmee import aasInMailFps
#from bisos.marmee import aasOutMailFps

# from bisos.qmail import maildrop
from bisos.qmail import qmail
from bisos.qmail import qmailControl

from bisos.common import lines

import enum


####+BEGIN: bx:cs:py3:section :title "Common Parameters Specification"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  /Section/    [[elisp:(outline-show-subtree+toggle)][||]] *Common Parameters Specification*  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:

####+BEGIN: bx:dblock:python:func :funcName "commonParamsSpecify" :funcType "ParSpec" :retType "" :deco "" :argsList "csParams"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  F-ParSpec  [[elisp:(outline-show-subtree+toggle)][||]] /commonParamsSpecify/ retType= argsList=(csParams)  [[elisp:(org-cycle)][| ]]
#+end_org """
def commonParamsSpecify(
    csParams,
):
####+END:
    csParams.parDictAdd(
        parName='qAddrAcct',
        parDescription="Qmail O/R Account",
        parDataType=None,
        parDefault=None,
        parChoices=["any"],
        # parScope=icm.CmndParamScope.TargetParam,
        argparseShortOpt=None,
        argparseLongOpt='--qAddrAcct',
    )
    csParams.parDictAdd(
        parName='localPart',
        parDescription="Qmail O/R Account Address",
        parDataType=None,
        parDefault=None,
        parChoices=["any"],
        # parScope=icm.CmndParamScope.TargetParam,
        argparseShortOpt=None,
        argparseLongOpt='--localPart',
    )
    csParams.parDictAdd(
        parName='qmailAcct',
        parDescription="Qmail O/R Account",
        parDataType=None,
        parDefault=None,
        parChoices=["any"],
        # parScope=icm.CmndParamScope.TargetParam,
        argparseShortOpt=None,
        argparseLongOpt='--qmailAcct',
    )
    csParams.parDictAdd(
        parName='qmailAddr',
        parDescription="Qmail O/R Account Address",
        parDataType=None,
        parDefault=None,
        parChoices=["any"],
        # parScope=icm.CmndParamScope.TargetParam,
        argparseShortOpt=None,
        argparseLongOpt='--qmailAddr',
    )
    csParams.parDictAdd(
        parName='maildir',
        parDescription="Directory path to Maildir",
        parDataType=None,
        parDefault=None,
        parChoices=["any"],
        # parScope=icm.CmndParamScope.TargetParam,
        argparseShortOpt=None,
        argparseLongOpt='--maildir',
    )


####+BEGIN: bx:cs:py3:section :title "CS-Lib Examples"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  /Section/    [[elisp:(outline-show-subtree+toggle)][||]] *CS-Lib Examples*  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:

####+BEGIN: b:py3:cs:func/typing :funcName "examples_csu" :funcType "eType" :retType "" :deco "default" :argsList ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  F-T-eType  [[elisp:(outline-show-subtree+toggle)][||]] /examples_csu/  deco=default  [[elisp:(org-cycle)][| ]]
#+end_org """
@cs.track(fnLoc=True, fnEntry=True, fnExit=True)
def examples_csu(
####+END:
        qmailAcct: str = "alias",
        qmailAddr: str = "postmaster",
        sectionTitle: typing.AnyStr = "",
) -> None:
    """ #+begin_org
** [[elisp:(org-cycle)][| *DocStr | ] Examples of Service Access Instance Commands.
    #+end_org """

    od = collections.OrderedDict
    cmnd = cs.examples.cmndEnter
    literal = cs.examples.execInsert

    includePath = qmail.installation.usersBaseDir.joinpath("include")
    assignPath = qmail.installation.usersBaseDir.joinpath("assign")

    dotQmailParams = od([('qAddrAcct', qmailAcct), ('localPart', qmailAddr),])

    #cmnd('cmdbSummary', pars=perfNamePars, comment=" # remote obtain facter data, use it to summarize for cmdb")

    cs.examples.menuChapter('*BxQmail Account And Addrs Utilities*')

    # cmnd('qmailAcctAddr_maildropUpdate', pars=acctAddr, comment=f" # Based on System config, control/me=")

    cs.examples.menuChapter('*LocalDeliveryAcct*')

    cmnd('localDeliveryAcct', args="ensureUsersBaseDir", comment=f" # create usersBaseDir={str(qmail.installation.usersBaseDir)} if needed")
    cmnd('localDeliveryAcct', args="newUserProc", comment=f" # Runs qmail-pw2u and  qmail-newu")
    cmnd('localDeliveryAcct', args="add alias", comment=f" # Adds user  to {includePath} and newUserProc")
    cmnd('localDeliveryAcct', args="add bisos", comment=f" # Adds user to {includePath} and newUserProc")
    cmnd('localDeliveryAcct', args="delete bisos", comment=f" # Deletes user from {includePath} and newUserProc")
    cmnd('localDeliveryAcct', args="verify bisos", comment=f" # Verifies user is in {assignPath}")
    cmnd('localDeliveryAcct', args="mainDomainGet", comment=f" # control/locals={qmailControl.QCFV_QmailSend().locals}")
    literal(f"ls -l {qmail.installation.usersBaseDir}/*")
    literal(f"find {qmail.installation.usersBaseDir} -type f -print | grep -v cdb  | xargs grep ^")

    cs.examples.menuChapter('*Maildir of qAddrAcct*')

    qAddrAcctMaildirParams = od([('qAddrAcct', qmailAcct), ('maildir', "./Maildir/"),])
    cmnd('maildir', args="create", pars=qAddrAcctMaildirParams, comment=f" # ")
    cmnd('maildir', args="delete", pars=qAddrAcctMaildirParams, comment=f" # ")
    cmnd('maildir', args="verify", pars=qAddrAcctMaildirParams, comment=f" # ")

    cs.examples.menuChapter('*DotQmail*')

    dotQmailMaildirParams = od([('qAddrAcct', qmailAcct), ('localPart', qmailAddr), ('maildir', "./Maildir/"),])

    cmnd('dotQmailFile', args="read", pars=dotQmailParams, comment=f" # ")
    cmnd('dotQmailFile', args="write", pars=dotQmailParams, comment=f" # Not Yet")
    cmnd('dotQmailFile', args="addMaildirLine", pars=dotQmailMaildirParams, comment=f" # ")
    cmnd('dotQmailFile', args="addForwardLine 'destEmailComesHere'", pars=dotQmailParams, comment=f" # ")
    cmnd('dotQmailFile', args="deleteLine 'destEmailComesHere'", pars=dotQmailParams, comment=f" # ")

    cs.examples.menuChapter('*VirDomEntry*')

    cmnd('virDomEntry', args="add", comment=f" # ")
    cmnd('virDomEntry', args="delete", comment=f" # ")
    cmnd('virDomEntry', args="acctGet", comment=f" # ")
    cmnd('virDomEntry', args="domainGet", comment=f" # ")

    cs.examples.menuChapter('*RcpthostsEntry*')

    cmnd('rcpthostsEntry', args="add", comment=f" # ")
    cmnd('rcpthostsEntry', args="delete", comment=f" # ")
    cmnd('rcpthostsEntry', args="acctGet", comment=f" # ")
    cmnd('rcpthostsEntry', args="domainGet", comment=f" # ")

    cs.examples.menuChapter('*VirDom*')

    cmnd('virDom', args="update", comment=f" # ")
    cmnd('virDom', args="delete", comment=f" # ")
    cmnd('virDom', args="acctGet", comment=f" # ")
    cmnd('virDom', args="domainGet", comment=f" # ")

    cs.examples.menuChapter('*Direct Commands*')

    literal("hostname --fqdn", comment=" # usually used as control/me")


####+BEGIN: bx:cs:py3:section :title "CS-Commands"
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  /Section/    [[elisp:(outline-show-subtree+toggle)][||]] *CS-Commands*  [[elisp:(org-cycle)][| ]]
#+end_org """
####+END:

####+BEGIN: b:py3:cs:cmnd/classHead :cmndName "qmailAcctAddr_maildropUpdate" :comment "" :extent "verify" :parsMand "qmailAcct qmailAddr" :argsMin 1 :argsMax 1 :pyInv ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  CmndSvc-   [[elisp:(outline-show-subtree+toggle)][||]] <<qmailAcctAddr_maildropUpdate>>  =verify= parsMand=qmailAcct qmailAddr argsMin=1 argsMax=1 ro=cli   [[elisp:(org-cycle)][| ]]
#+end_org """
class qmailAcctAddr_maildropUpdate(cs.Cmnd):
    cmndParamsMandatory = [ 'qmailAcct', 'qmailAddr', ]
    cmndParamsOptional = [ ]
    cmndArgsLen = {'Min': 1, 'Max': 1,}

    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmnd(self,
             rtInv: cs.RtInvoker,
             cmndOutcome: b.op.Outcome,
             qmailAcct: typing.Optional[str]=None,  # Cs Mandatory Param
             qmailAddr: typing.Optional[str]=None,  # Cs Mandatory Param
             argsList: typing.Optional[list[str]]=None,  # CsArgs
    ) -> b.op.Outcome:

        callParamsDict = {'qmailAcct': qmailAcct, 'qmailAddr': qmailAddr, }
        if self.invocationValidate(rtInv, cmndOutcome, callParamsDict, argsList).isProblematic():
            return b_io.eh.badOutcome(cmndOutcome)
        cmndArgsSpecDict = self.cmndArgsSpec()
        qmailAcct = csParam.mappedValue('qmailAcct', qmailAcct)
        qmailAddr = csParam.mappedValue('qmailAddr', qmailAddr)
####+END:
        self.cmndDocStr(f""" #+begin_org
** [[elisp:(org-cycle)][| *CmndDesc:* | ]] For dotQmail specified by qmailAcct and qmailAddr, update maildrop to maildropQmailAddr.
*** Make sure that file corresponing to =maildropQmailAddr= exists.
*** Read in qmailAcctAddr.
*** Update maildrop line in qmailAcctAddr file.
*** -
        #+end_org """)

        cmndArgsSpecDict = self.cmndArgsSpec()
        maildropQmailAddr = self.cmndArgsGet("0", cmndArgsSpecDict, argsList)

        qmailAcctAddr = qmail.AcctAddr(qmailAcct, qmailAddr)
        qmailAcctPath = qmailAcctAddr.acctPath()

        maildropRelPath = maildrop.maildropFileName(qmailAddr)
        maildropPath = qmailAcctPath.joinpath(maildropRelPath)

        if not maildropPath.exists():
            print(f"Missing {maildropPath}")
            return b_io.eh.badOutcome(cmndOutcome)

        maildropLine = f"""| maildrop {maildropRelPath}"""

        dotQmailContent = qmailAcctAddr.dotQmailFileContentRead()
        dotQmailLinesObj = lines.Lines(inContent=dotQmailContent)
        updatedDotQmailLines = dotQmailLinesObj.addIfNotThere(maildropLine)

        updatedDotQmailContent = '\n'.join(updatedDotQmailLines)
        qmailAcctAddr.dotQmailFileContentWrite(updatedDotQmailContent)

        return(cmndOutcome)


####+BEGIN: b:py3:cs:method/args :methodName "cmndArgsSpec" :methodType "anyOrNone" :retType "bool" :deco "default" :argsList "self"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-anyOrNone [[elisp:(outline-show-subtree+toggle)][||]] /cmndArgsSpec/ deco=default  deco=default   [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmndArgsSpec(self, ):
####+END:
        """  #+begin_org
** [[elisp:(org-cycle)][| *cmndArgsSpec:* | ]]
        #+end_org """

        cmndArgsSpecDict = cs.arg.CmndArgsSpecDict()
        cmndArgsSpecDict.argsDictAdd(
            argPosition="0",
            argName="maildropQmailAddr",
            argChoices=[],
            argDescription="Maildrop File Identifier"
        )
        return cmndArgsSpecDict


####+BEGIN: b:py3:cs:cmnd/classHead :cmndName "localDeliveryAcct" :comment "" :extent "verify" :parsMand "" :argsMin 1 :argsMax 4 :pyInv ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  CmndSvc-   [[elisp:(outline-show-subtree+toggle)][||]] <<localDeliveryAcct>>  =verify= argsMin=1 argsMax=4 ro=cli   [[elisp:(org-cycle)][| ]]
#+end_org """
class localDeliveryAcct(cs.Cmnd):
    cmndParamsMandatory = [ ]
    cmndParamsOptional = [ ]
    cmndArgsLen = {'Min': 1, 'Max': 4,}

    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmnd(self,
             rtInv: cs.RtInvoker,
             cmndOutcome: b.op.Outcome,
             argsList: typing.Optional[list[str]]=None,  # CsArgs
    ) -> b.op.Outcome:

        failed = b_io.eh.badOutcome
        callParamsDict = {}
        if self.invocationValidate(rtInv, cmndOutcome, callParamsDict, argsList).isProblematic():
            return failed(cmndOutcome)
        cmndArgsSpecDict = self.cmndArgsSpec()
####+END:
        self.cmndDocStr(f""" #+begin_org
** [[elisp:(org-cycle)][| *CmndDesc:* | ]] For dotQmail specified by qmailAcct and qmailAddr, update maildrop to maildropQmailAddr.
*** Make sure that file corresponing to =maildropQmailAddr= exists.
*** Read in qmailAcctAddr.
*** Update maildrop line in qmailAcctAddr file.
*** -
        #+end_org """)

        cmndArg = self.cmndArgsGet("0", cmndArgsSpecDict, argsList)
        restArgs = self.cmndArgsGet("1&9999", cmndArgsSpecDict, argsList)

        result: typing.Any = None

        if cmndArg == 'ensureUsersBaseDir':
            cmndOutcome = qmail.LocalDeliveryAcct.ensureUsersBaseDir()
        elif cmndArg == 'newUserProc':
            cmndOutcome = qmail.LocalDeliveryAcct.newUserProc()
        elif cmndArg == 'add':
            cmndOutcome = qmail.LocalDeliveryAcct.add(restArgs[0])
        elif cmndArg == 'delete':
            cmndOutcome = qmail.LocalDeliveryAcct.delete(restArgs[0])
        elif cmndArg == 'verify':
            cmndOutcome = qmail.LocalDeliveryAcct.verify(restArgs[0])
        elif cmndArg == 'mainDomainGet':
            cmndOutcome = qmail.LocalDeliveryAcct.mainDomainGet()
        else:
            b_io.eh.critical_usageError(f"Unknown cmndArg={cmndArg}")

        return(cmndOutcome)


####+BEGIN: b:py3:cs:method/args :methodName "cmndArgsSpec" :methodType "anyOrNone" :retType "bool" :deco "default" :argsList "self"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-anyOrNone [[elisp:(outline-show-subtree+toggle)][||]] /cmndArgsSpec/ deco=default  deco=default   [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmndArgsSpec(self, ):
####+END:
        """  #+begin_org
** [[elisp:(org-cycle)][| *cmndArgsSpec:* | ]]
        #+end_org """

        cmndArgsSpecDict = cs.arg.CmndArgsSpecDict()
        cmndArgsSpecDict.argsDictAdd(
            argPosition="0",
            argName="cmndArg",
            argChoices=[],
            argDescription="Command which may need restOfArgs"
        )
        cmndArgsSpecDict.argsDictAdd(
            argPosition="1&9999",
            argName="restOfArgs",
            argChoices=[],
            argDescription="Rest of args which may be specific to each command"
        )
        return cmndArgsSpecDict


    ####+BEGIN: b:py3:cs:cmnd/classHead :cmndName "maildir" :comment "" :extent "verify" :parsMand "qAddrAcct" :argsMin 1 :argsMax 9999 :pyInv ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  CmndSvc-   [[elisp:(outline-show-subtree+toggle)][||]] <<maildir>>  =verify= parsMand=qAddrAcct argsMin=1 argsMax=9999 ro=cli   [[elisp:(org-cycle)][| ]]
#+end_org """
class maildir(cs.Cmnd):
    cmndParamsMandatory = [ 'qAddrAcct', ]
    cmndParamsOptional = [ ]
    cmndArgsLen = {'Min': 1, 'Max': 9999,}

    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmnd(self,
             rtInv: cs.RtInvoker,
             cmndOutcome: b.op.Outcome,
             qAddrAcct: typing.Optional[str]=None,  # Cs Mandatory Param
             argsList: typing.Optional[list[str]]=None,  # CsArgs
    ) -> b.op.Outcome:

        failed = b_io.eh.badOutcome
        callParamsDict = {'qAddrAcct': qAddrAcct, }
        if self.invocationValidate(rtInv, cmndOutcome, callParamsDict, argsList).isProblematic():
            return failed(cmndOutcome)
        cmndArgsSpecDict = self.cmndArgsSpec()
        qAddrAcct = csParam.mappedValue('qAddrAcct', qAddrAcct)
####+END:
        self.cmndDocStr(f""" #+begin_org
** [[elisp:(org-cycle)][| *CmndDesc:* | ]] For dotQmail specified by qmailAcct and qmailAddr, update maildrop to maildropQmailAddr.
*** Make sure that file corresponing to =maildropQmailAddr= exists.
*** Read in qmailAcctAddr.
*** Update maildrop line in qmailAcctAddr file.
*** -
        #+end_org """)

        cmndArg = self.cmndArgsGet("0", cmndArgsSpecDict, argsList)
        restArgs = self.cmndArgsGet("1&9999", cmndArgsSpecDict, argsList)

        result: typing.Any = None

        if cmndArg == 'create':
            cmndOutcome = qmail.Maildir.create(qAddrAcct, restArgs)
        elif cmndArg == 'delete':
            cmndOutcome = qmail.Maildir.delete(qAddrAcct, restArgs)
        elif cmndArg == 'verify':
            cmndOutcome = qmail.Maildir.verify(qAddrAcct, restArgs)
        else:
            b_io.eh.critical_usageError(f"Unknown cmndArg={cmndArg}")

        return(cmndOutcome)


####+BEGIN: b:py3:cs:method/args :methodName "cmndArgsSpec" :methodType "anyOrNone" :retType "bool" :deco "default" :argsList "self"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-anyOrNone [[elisp:(outline-show-subtree+toggle)][||]] /cmndArgsSpec/ deco=default  deco=default   [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmndArgsSpec(self, ):
####+END:
        """  #+begin_org
** [[elisp:(org-cycle)][| *cmndArgsSpec:* | ]]
        #+end_org """

        cmndArgsSpecDict = cs.arg.CmndArgsSpecDict()
        cmndArgsSpecDict.argsDictAdd(
            argPosition="0",
            argName="cmndArg",
            argChoices=[],
            argDescription="Command which may need restOfArgs"
        )
        cmndArgsSpecDict.argsDictAdd(
            argPosition="1&9999",
            argName="restOfArgs",
            argChoices=[],
            argDescription="Rest of args which may be specific to each command"
        )
        return cmndArgsSpecDict

####+BEGIN: b:py3:cs:cmnd/classHead :cmndName "dotQmailFile" :comment "" :extent "verify" :parsMand "qAddrAcct localPart" :parsOpt "maildir" :argsMin 1 :argsMax 4 :pyInv ""
""" #+begin_org
*  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  CmndSvc-   [[elisp:(outline-show-subtree+toggle)][||]] <<dotQmailFile>>  =verify= parsMand=qAddrAcct localPart parsOpt=maildir argsMin=1 argsMax=4 ro=cli   [[elisp:(org-cycle)][| ]]
#+end_org """
class dotQmailFile(cs.Cmnd):
    cmndParamsMandatory = [ 'qAddrAcct', 'localPart', ]
    cmndParamsOptional = [ 'maildir', ]
    cmndArgsLen = {'Min': 1, 'Max': 4,}

    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmnd(self,
             rtInv: cs.RtInvoker,
             cmndOutcome: b.op.Outcome,
             qAddrAcct: typing.Optional[str]=None,  # Cs Mandatory Param
             localPart: typing.Optional[str]=None,  # Cs Mandatory Param
             maildir: typing.Optional[str]=None,  # Cs Optional Param
             argsList: typing.Optional[list[str]]=None,  # CsArgs
    ) -> b.op.Outcome:

        failed = b_io.eh.badOutcome
        callParamsDict = {'qAddrAcct': qAddrAcct, 'localPart': localPart, 'maildir': maildir, }
        if self.invocationValidate(rtInv, cmndOutcome, callParamsDict, argsList).isProblematic():
            return failed(cmndOutcome)
        cmndArgsSpecDict = self.cmndArgsSpec()
        qAddrAcct = csParam.mappedValue('qAddrAcct', qAddrAcct)
        localPart = csParam.mappedValue('localPart', localPart)
        maildir = csParam.mappedValue('maildir', maildir)
####+END:
        self.cmndDocStr(f""" #+begin_org
** [[elisp:(org-cycle)][| *CmndDesc:* | ]] For dotQmail specified by qmailAcct and qmailAddr, update maildrop to maildropQmailAddr.
*** Make sure that file corresponing to =maildropQmailAddr= exists.
*** Read in qmailAcctAddr.
*** Update maildrop line in qmailAcctAddr file.
*** -
        #+end_org """)

        cmndArg = self.cmndArgsGet("0", cmndArgsSpecDict, argsList)
        restArgs = self.cmndArgsGet("1&9999", cmndArgsSpecDict, argsList)

        result: typing.Any = None

        dotQmail = qmail.DotQmailFile(qAddrAcct, localPart)

        if cmndArg == 'read':
            result = dotQmail.contentRead()
            cmndOutcome.set(opResults=result)
        elif cmndArg == 'addMaildirLine':
            cmndOutcome = dotQmail.addMaildirLine(maildir, cmndOutcome)
        elif cmndArg == 'addForwardLine':
            cmndOutcome = dotQmail.addForwardLine(restArgs[0], cmndOutcome)
        elif cmndArg == 'deleteLine':
            cmndOutcome = dotQmail.deleteLine(restArgs[0], cmndOutcome)
        else:
            b_io.eh.critical_usageError(f"Unknown cmndArg={cmndArg}")

        return(cmndOutcome)


####+BEGIN: b:py3:cs:method/args :methodName "cmndArgsSpec" :methodType "anyOrNone" :retType "bool" :deco "default" :argsList "self"
    """ #+begin_org
**  _[[elisp:(blee:menu-sel:outline:popupMenu)][±]]_ _[[elisp:(blee:menu-sel:navigation:popupMenu)][Ξ]]_ [[elisp:(outline-show-branches+toggle)][|=]] [[elisp:(bx:orgm:indirectBufOther)][|>]] *[[elisp:(blee:ppmm:org-mode-toggle)][|N]]*  Mtd-T-anyOrNone [[elisp:(outline-show-subtree+toggle)][||]] /cmndArgsSpec/ deco=default  deco=default   [[elisp:(org-cycle)][| ]]
    #+end_org """
    @cs.track(fnLoc=True, fnEntry=True, fnExit=True)
    def cmndArgsSpec(self, ):
####+END:
        """  #+begin_org
** [[elisp:(org-cycle)][| *cmndArgsSpec:* | ]]
        #+end_org """

        cmndArgsSpecDict = cs.arg.CmndArgsSpecDict()
        cmndArgsSpecDict.argsDictAdd(
            argPosition="0",
            argName="cmndArg",
            argChoices=[],
            argDescription="Command which may need restOfArgs"
        )
        cmndArgsSpecDict.argsDictAdd(
            argPosition="1&9999",
            argName="restOfArgs",
            argChoices=[],
            argDescription="Rest of args which may be specific to each command"
        )
        return cmndArgsSpecDict



####+BEGIN: b:py3:cs:framework/endOfFile :basedOn "classification"
""" #+begin_org
* [[elisp:(org-cycle)][| *End-Of-Editable-Text* |]] :: emacs and org variables and control parameters
#+end_org """

#+STARTUP: showall

### local variables:
### no-byte-compile: t
### end:
####+END:
